package se.chalmers.paragon;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;

/**
 * <p>
 * Implements a Lock, providing a factory method to construct new lock
 * instances. Handles the opening and closing of locks, ensures that correct
 * arity is maintained via the {@link se.chalmers.paragon.LockState} class.
 * </p>
 * 
 * @author Paragon Developers (http://www.cse.chalmers.se/paragon)
 * 
 */
public class Lock {

	/**
	 * Global clauses on Flow
	 */
	// public static final Lock FLOW = Lock.newLock("Flow", 1);

	/**
	 * Lists all the clauses that have this lock instance as a head.
	 */
	private LinkedList<Clause> relatedClauses = new LinkedList<Clause>();

	/**
	 * The number of arguments for this lock instance.
	 */
	private int arity;

	/**
	 * Store the explicitly opened / closed locks
	 */
	private LockState explicit_lockState;

	/**
	 * Store derived open locks from clauses
	 */
	private LockState implicit_lockState;

	/**
	 * The name of this lock (for messages)
	 */
	private String lockName;

	/**
	 * For storing 'assumption facts' used in the implies-method
	 */
	protected LockState freshFacts_lockState;

	/**
	 * Keeping track of all registered locks
	 */
	private static HashSet<String> existingNames = new HashSet<String>();

	/**
	 * Set to true when rules on Lock.FLOW need to be taken into account.
	 */
	private boolean addGlobalFlowClauses = false;

	/**
	 * Set to false whenever it becomes impossible to change the global rules
	 * (i.e. after some derivation has taken place)
	 */
	private boolean canChangeClauses = true;

	/**
	 * Set to true when lock is a FLOW lock.
	 */
	private boolean isFlowLock = false;

	/**
	 * Constructs a new lock
	 * 
	 * @param lockName
	 *            The name of this lock
	 * @param arity
	 *            The number of arguments for this lock
	 */
	protected Lock(String lockName, int arity) {
		if (existingNames == null)
			existingNames = new HashSet<String>();
		if (existingNames.contains(lockName))
			throw new UniqueNameError("Lock name " + lockName
					+ " is not unique.");
		existingNames.add(lockName);
		this.lockName = lockName;
		this.arity = arity;
		this.explicit_lockState = new LockState(this);
		this.implicit_lockState = new LockState(this);
		this.freshFacts_lockState = new LockState(this);
	}

	/**
	 * Returns a new lock
	 * 
	 * @param lockName
	 *            Name of the lock
	 * @param arity
	 *            Number of arguments of the lock
	 * @return
	 */
	public static Lock newLock(String lockName, int arity) {
		return new Lock(lockName, arity);
	}

	/**
	 * Returns the special reserved isActor lock
	 */
	protected static Lock ISACTOR = newLock("isActor", 1);

	/**
	 * 
	 * @return The number of arguments that this lock takes
	 */
	public int getArity() {
		return this.arity;
	}

	/**
	 * The name of this lock
	 * 
	 * @return
	 */
	public String getName() {
		return this.lockName;
	}

	public String getNameString() {
		return this.isFlowLock ? "Flow" : this.lockName;
	}

	public void addClause(ActorList head, Atom... body) {
		if (!canChangeClauses)
			throw new LockException("Trying to add clause to "
					+ (isFlowLock ? "policy " : "lock ") + getName()
					+ " after it has been used in some derivation.");
		this.relatedClauses.add(Clause
				.newClause(Atom.newAtom(this, head), body));
	}

	public void transitive() {
		this.addClause(ActorList.newActorList(Actor.newActorVariable(0), Actor
				.newActorVariable(2)), Atom.newAtom(this, Actor
				.newActorVariable(0), Actor.newActorVariable(1)), Atom.newAtom(
				this, Actor.newActorVariable(1), Actor.newActorVariable(2)));
	}

	public void symmetric() {
		this.addClause(ActorList.newActorList(Actor.newActorVariable(0), Actor
				.newActorVariable(1)), Atom.newAtom(this, Actor
				.newActorVariable(1), Actor.newActorVariable(0)));
	}

	public void reflexive() {
		this.addClause(ActorList.newActorList(Actor.newActorVariable(0), Actor
				.newActorVariable(0)));
	}

	/**
	 * Open a lock explicitly for the provided arguments
	 * 
	 * @param actors
	 *            List of concrete actors for which this lock should be openened
	 * 
	 * @return true if the lock was not already opened
	 */
	public boolean open(ConcreteActor... actors) {
		return explicit_lockState.open(actors);
	}

	/**
	 * Close a lock explicitly for the provided arguments
	 * 
	 * @param actors
	 *            List of concrete actors for which this lock should be closed
	 */
	public boolean close(ConcreteActor... actors) {
		return explicit_lockState.close(actors);
	}

	/**
	 * @deprecated
	 * @param actors
	 *            List of concrete actors to check the <i>explicit</i> state of
	 *            this lock for
	 * @return Returns true if this lock has been explicitly opened.
	 */
	public boolean isExplicitOpen(ConcreteActor... actors) {
		return explicit_lockState.isOpen(actors);
	}

	/**
	 * Checks whether the lock is open for the provided arguments. When the
	 * explicit state does not have this lock registered as open, relevant lock
	 * clauses are applied the implicit state until either
	 * <ul>
	 * <li>The lock is derived to be open for the arguments
	 * <li>The entire implicit state is derived and the lock is not open for
	 * this instance
	 * </ul>
	 * 
	 * @param actors
	 *            List of concrete actors to check the state of this lock for
	 * @return Returns true if this lock is open, either explicit or via clauses
	 */
	public boolean isOpen(ConcreteActor... actors) {
		// Is the lock explicitly opened?
		if (explicit_lockState.isOpen(actors)
				|| freshFacts_lockState.isOpen(actors))
			return true;
		// Derive the implicit state
		boolean newFactsDerived;
		do {
			// Try to derive new facts for one iteration
			newFactsDerived = deduceOneStep(new HashSet<Lock>());
			// If the open-status is derived, stop deriving further, reset the
			// implicit state and return that the lock is open.
			if (implicit_lockState.isOpen(actors)) {
				cleanUpImplicit(new HashSet<Lock>());
				return true;
			}
			// Continue as long as new facts were derived; as long as new locks
			// are opened which are of a lock relevant to this lock, we might
			// derive something for this lock in a later iteration.
		} while (newFactsDerived);
		// No new facts were derived, thus we will derive nothing new in a next
		// iteration either: the requested lock is not opened.
		cleanUpImplicit(new HashSet<Lock>());
		return false;
	}

	/**
	 * For all relevant clauses, i.e. clauses with this lock as a head, try to
	 * apply them and derive a new fact on this lock.
	 * 
	 * @return True if new facts were derived, false otherwise
	 */
	protected boolean deduceOneStep(HashSet<Lock> locksIterated) {
		locksIterated.add(this);
		boolean newFacts = false;
		for (Clause clause : getRelatedClauses()) {
			newFacts = newFacts
					|| clause
							.deduceOneStep(new HashMap<ActorVariable, ConcreteActor>());
			newFacts = newFacts
					|| clause.deduceRelevantLocksOneStep(locksIterated);
		}
		return newFacts;
	}

	/**
	 * 
	 * @param actors
	 *            The actors to check for
	 * @return True if this lock instance is explicitly open or has been derived
	 *         to be open
	 */
	protected boolean isExplicitOrImplicitOpen(ConcreteActor... actors) {
		return explicit_lockState.isOpen(actors)
				|| freshFacts_lockState.isOpen(actors)
				|| implicit_lockState.isOpen(actors);
	}

	/**
	 * Empties the state of implicit derived locks, calls the same function on
	 * locks that have contributed to the derivation of this state.
	 * 
	 * @param locksIterated
	 *            Keeps track of the iterated locks to prevent infinite loops
	 */
	protected void cleanUpImplicit(HashSet<Lock> locksIterated) {
		this.implicit_lockState = new LockState(this);
		locksIterated.add(this);
		for (Clause clause : getRelatedClauses()) {
			clause.cleanUpImplicit(locksIterated);
		}
	}

	/**
	 * 
	 * @param actors
	 *            List of actors to add to the state of implicit (derived) open
	 *            locks.
	 * @return True if this lock was not already open, false otherwise
	 */
	protected boolean openImplicit(ConcreteActor... actors) {
		if (explicit_lockState.isOpen(actors)
				|| freshFacts_lockState.isOpen(actors))
			return false;
		return implicit_lockState.open(actors);
	}

	/**
	 * Returns a set of actor pairs that agree with the provided array of actors
	 * on the concrete actors, and are opened in either the explicit or implicit
	 * lock state.
	 * 
	 * @param actors
	 *            Array of actors, containing at least one actor variable
	 * @return (Possibly empty) set of actor pairs which have the same concrete
	 *         actor the same position as in the provided argument and open in
	 *         the explicit or implicit lock state of this lock.
	 */
	protected HashSet<ActorList> getAllOpen(Actor[] actors) {
		HashSet<ActorList> res_aps = new HashSet<ActorList>();
		// Matching open locks in the explicit state
		for (ActorList ap : explicit_lockState.currentState) {
			boolean possible = true;
			for (int i = 0; i < actors.length; i++)
				if (actors[i] instanceof ConcreteActor)
					if (!ap.getActors()[i].equals(actors[i])) {
						possible = false;
						break;
					}
			if (possible)
				res_aps.add(ap);
		}
		// Matching open locks in the fresh state
		for (ActorList ap : freshFacts_lockState.currentState) {
			boolean possible = true;
			for (int i = 0; i < actors.length; i++)
				if (actors[i] instanceof ConcreteActor)
					if (!ap.getActors()[i].equals(actors[i])) {
						possible = false;
						break;
					}
			if (possible)
				res_aps.add(ap);
		}
		// Matching open locks in the implicit state
		for (ActorList ap : implicit_lockState.currentState) {
			boolean possible = true;
			for (int i = 0; i < actors.length; i++)
				if (actors[i] instanceof ConcreteActor)
					if (!ap.getActors()[i].equals(actors[i])) {
						possible = false;
						break;
					}
			if (possible)
				res_aps.add(ap);
		}
		return res_aps;
	}

	protected LinkedList<Clause> getRelatedClauses() {
		if (addGlobalFlowClauses) {
			// Removed, was for adding the global policy.
			// this.relatedClauses.addAll(Util.copyClause(
			// FLOW.getRelatedClauses(), this));
			this.addGlobalFlowClauses = false; // only once
		}
		this.canChangeClauses = false;
		return relatedClauses;
	}

	protected void openFreshFact(ConcreteActor[] actors) {
		this.freshFacts_lockState.open(actors);
	}

	protected void clearFreshFacts() {
		this.freshFacts_lockState = new LockState(this);
	}

	public String toString() {
		String res = "";
		if (this.isFlowLock)
			res += "Flow";
		else
			res += getName();
		res += " {\n";
		for (Clause r : getRelatedClauses())
			res += "   " + r.toString() + "\n";
		res += "  }";
		return res;
	}

	public static Lock freshFlowLock(String base) {
		int i = 0;
		String name = base;
		while (existingNames.contains(name)) {
			name = base + i;
			i += 1;
		}
		Lock flow = new Lock(name, 1);
		flow.isFlowLock = true;
		flow.addGlobalFlowClauses = true;
		return flow;
	}

	public boolean isFlowLock() {
		return this.isFlowLock;
	}

}
